package lsedit;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Event;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import javax.swing.undo.*;
import javax.swing.event.UndoableEditEvent;
import javax.swing.event.UndoableEditListener;

public class HistoryBox extends JComponent implements ToolBarEventHandler, MouseListener, MouseMotionListener
{

	public class ShowCompoundEdit extends JDialog { 

		public class ShowCompoundEditPanel extends JComponent {
			
			Vector m_v;

			ShowCompoundEditPanel(Vector v)
			{
				m_v = v;
			}

			public void paintComponent(Graphics g)
			{
				int height = HistoryBox.HEIGHT;
				Vector v   = m_v;

				setBackground(Color.lightGray);

				if (v != null) {
					Enumeration		en;
					UndoableEdit	undoableEdit;

					g.setFont(m_textFont);
					g.setColor(Color.BLACK);

					for (en = v.elements(); en.hasMoreElements(); ) {
						undoableEdit = (UndoableEdit) en.nextElement();
						if (undoableEdit instanceof MyPaintableUndoableEdit) {
							((MyPaintableUndoableEdit) undoableEdit).paintComponent(g, MARGIN, height);
						} else {
							g.drawString(undoableEdit.getPresentationName(), MARGIN, height+BASELINE);
						}
						height      += HistoryBox.HEIGHT;
			}	}	}
		}

		MyCompoundEdit		m_compoundEdit;

		public ShowCompoundEdit(HistoryBox historyBox, MyCompoundEdit compoundEdit)
		{
			super(historyBox.m_ls.getFrame(), compoundEdit.getPresentationName());

			int				x, y, width, height, w;
			Enumeration		en;
			UndoableEdit	undoableEdit;
			Vector			v;
			Dimension		preferredSize = new Dimension();
			Container		contentPane;
			ShowCompoundEditPanel panel;

			x      = historyBox.m_tabbedPane.getX() + 20;
			y      = historyBox.m_tabbedPane.getY() + 20;
			setLocation(x,y);

			width  = historyBox.getWidth();
			height = historyBox.getHeight();
			m_compoundEdit = compoundEdit;
			v              = compoundEdit.getEdits();

			preferredSize.height = (v.size() * HistoryBox.HEIGHT) + HistoryBox.GAP + 20;
			preferredSize.width  = 0;
			for (en = v.elements(); en.hasMoreElements(); ) {
				undoableEdit = (UndoableEdit) en.nextElement();

				if (undoableEdit instanceof MyPaintableUndoableEdit) {
					w   = ((MyPaintableUndoableEdit) undoableEdit).getPreferredWidth();
				} else {
					FontMetrics	fm;

					fm   = getFontMetrics(HistoryBox.m_textFont);
					w    = fm.stringWidth(undoableEdit.getPresentationName());
				}
				if (w > preferredSize.width) {
					preferredSize.width = w;
			}	}
			preferredSize.width  += HistoryBox.GAP;
			if (preferredSize.width < width / 2) {
				preferredSize.width = width / 2;
			}
			contentPane = getContentPane();
			contentPane.setLayout(new BorderLayout());
			contentPane.setSize(preferredSize);
			panel       = new ShowCompoundEditPanel(v);
						panel.setPreferredSize(preferredSize);
			panel.setSize(preferredSize);
			contentPane.add(panel, BorderLayout.CENTER);

			pack();
			show();
		}
	}

	protected static final int MARGIN    = 5;
	protected static final int HEIGHT    = 16;
	protected static final int BASELINE  = 12;
	protected static final int GAP       = 5; 

	protected static final int TY_CLEAR	 = 0;
	protected static final int TY_RAISED = 1;
	protected static final int TY_SUNK	 = 2;

	protected LandscapeViewerCore   m_ls;
	protected JTabbedPane			m_tabbedPane;
	protected JScrollPane			m_scrollPane;

	public static Dimension         m_preferredSize    = new Dimension(0,0);

	public    static final String m_helpStr	 = "This box shows the history of updates made in the current diagram.";

	protected int					m_style;
	protected int					m_over;

	protected static final Font		m_textFont	 = new Font("Helvetica", Font.PLAIN, 11); 

	// --------------
	// Public methods 
	// --------------

	public HistoryBox(LandscapeViewerCore ls, JTabbedPane tabbedPane) 
	{
		m_ls          = ls;
		m_tabbedPane  = tabbedPane;

		setBackground(Diagram.boxColour);
//		setToolTipText(m_helpStr);

		m_scrollPane = new JScrollPane();
		m_scrollPane.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED); 
		m_scrollPane.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED); 
		m_scrollPane.setViewportView(this);
		tabbedPane.addTab("History", null, m_scrollPane, m_helpStr);
		addMouseListener(this);
		addMouseMotionListener(this);
	}

	public Dimension getPreferredSize()
	{
		return(m_preferredSize);
	}

	public LandscapeViewerCore getLegendLs()
	{
		return(m_ls);
	}

	public void fillHistoryBox()
	{
		repaint();
	}

	public void activate() 
	{
		fillHistoryBox();
		m_tabbedPane.setSelectedComponent(m_scrollPane);
	}

	public boolean isActive() 
	{
		Component active;

		if (!isVisible()) {
			return(false);
		}
		active = m_tabbedPane.getSelectedComponent();
		return(active != null && m_scrollPane == active);
	}

	public Dimension getMaximumSize()
	{
		return(getPreferredSize());
	}

	public Dimension getSize()
	{
		return(getPreferredSize());
	}


	public void paintComponent(Graphics g)
	{
		int height = HEIGHT;
		Vector v   = m_ls.getEdits();

		if (v != null) {
			Enumeration		en;
			UndoableEdit	undoableEdit;
			UndoableEdit	editToBeRedone = m_ls.getEditToBeRedone();
			String			name;
			int				at     = m_over;

			g.setFont(m_textFont);
			g.setColor(Color.BLACK);

			for (en = v.elements(); en.hasMoreElements(); --at) {
				undoableEdit = (UndoableEdit) en.nextElement();
				if (undoableEdit == editToBeRedone) {
					g.setColor(Color.RED);
				}
				if (at == 0) {
					if (m_style != MapBox.TY_CLEAR) {
						Color color;

						color = g.getColor();
						g.setColor(Color.BLACK);
						g.draw3DRect(0, height, m_preferredSize.width-1, HEIGHT-2, m_style == TY_RAISED);
						g.setColor(color);
				}	} 
				if (undoableEdit instanceof MyPaintableUndoableEdit) {
					((MyPaintableUndoableEdit) undoableEdit).paintComponent(g, MARGIN, height);
				} else if (undoableEdit instanceof MyCompoundEdit) {
					Color color;

					color = g.getColor();
					g.setColor(Color.WHITE);
					g.fill3DRect(0, height, m_preferredSize.width-1, HEIGHT-2, (at == 0 && m_style == TY_RAISED));
					g.setColor(color);
					g.drawString(undoableEdit.getPresentationName(), MARGIN, height+BASELINE);
				} else {
					g.drawString(undoableEdit.getPresentationName(), MARGIN, height+BASELINE);
				}
				height      += HEIGHT;
	}	}	}


	protected void mouseAt(int y)
	{
		int	over = (y / HEIGHT) - 1;
		int	max;
		int style;

		if (over < 0) {
			style = HistoryBox.TY_CLEAR;
			over  = -1;
		} else if (over >= (max = m_ls.countEdits())) {
			style = HistoryBox.TY_CLEAR;
			over  = max;
		} else {
			style = HistoryBox.TY_RAISED;
		}
		if (style != m_style || over != m_over) {
			m_style = style;
			m_over  = over;
			repaint();
	}	}

	// MouseListener interface

	public void mouseClicked(MouseEvent ev)
	{
	}

	public void mouseEntered(MouseEvent ev)
	{
		mouseAt(ev.getY());
	}

	public void mouseExited(MouseEvent ev)
	{
		m_style = MapBox.TY_CLEAR;
		repaint();
	}

	public void mousePressed(MouseEvent ev)
	{
		int over;

		m_over  = over = (ev.getY() / HEIGHT) - 1;
		if (ev.isMetaDown()) {
			Vector			v = m_ls.getEdits();
			if (v != null) {
				int				max;

				max = v.size();
				if (max > 0) {

					if (over >= 0 && over < max) {
						UndoableEdit	undoableEdit = (UndoableEdit) v.elementAt(over);

						if (undoableEdit instanceof MyCompoundEdit) {
							MyMenuItem	m1;
							JPopupMenu	m            = new JPopupMenu("Undo/Redo options");

							m1 = new MyMenuItem(m, "Show details", this, 0, Do.DESCEND);
							m1.setObject(undoableEdit);
							m.show(this, ev.getX(), ev.getY());
						}
					} else {
						JPopupMenu		m            = new JPopupMenu("Discard History");
						new MyMenuItem(m, "Dispose", this, 0, Do.DELETE);
						m.setForeground(Color.RED);
						m.show(this, ev.getX(), ev.getY());
			}	}	}
			return;
		}

		m_style = MapBox.TY_SUNK;
		if (over >= 0) {
			Vector v   = m_ls.getEdits();

			if (v != null) {
				Enumeration		en;
				UndoableEdit	undoableEdit;
				UndoableEdit	editToBeRedone = m_ls.getEditToBeRedone();
				String			name;
				int				at     = m_over;
				boolean			redo   = false;

				for (en = v.elements(); en.hasMoreElements(); --at) {
					undoableEdit = (UndoableEdit) en.nextElement();
					if (undoableEdit == editToBeRedone) {
						redo = true;
					}
					if (at == 0) {
						m_ls.massChange(undoableEdit, redo);
						break;
		}	}	}	}
		repaint();
	}

	public void mouseReleased(MouseEvent ev)
	{

		m_style = MapBox.TY_CLEAR;
		repaint();
	}

	// MouseMotionListener interface

	public void mouseDragged(MouseEvent ev)
	{
		mouseAt(ev.getY());

	}

	public void mouseMoved(MouseEvent ev)
	{
		mouseAt(ev.getY());
	}

	// ToolBarEventHandler

	 public void processKeyEvent(int key, int modifiers, Object object)
	 {
		switch (key) {
		case Do.DESCEND:
			new ShowCompoundEdit(this, (MyCompoundEdit) object);
			break;
		case Do.DELETE:
			m_ls.discardAllEdits();
			break;
		}
	 }

	 public void showInfo(String msg)
	 {
	 }		

}



